"""Package for exception handling."""

from .exceptions import (
    InvalidPasswordException,
    ProjectNotFoundException,
    UnexpectedDataError,
    UnexpectedFileContent,
    XknxProjectException,
)

__all__ = [
    "InvalidPasswordException",
    "ProjectNotFoundException",
    "UnexpectedDataError",
    "UnexpectedFileContent",
    "XknxProjectException",
]



"""Test utilities."""

from __future__ import annotations

from typing import Any

import pytest

from xknxproject import util
from xknxproject.models import DPTType, ParameterInstanceRef


@pytest.mark.parametrize(
    ("dpt_string", "expected"),
    [
        ("DPT-1", {"main": 1, "sub": None}),
        ("DPT-1 DPST-1-1", {"main": 1, "sub": None}),
        ("DPT-7 DPST-7-1", {"main": 7, "sub": None}),
        ("DPST-5-1", {"main": 5, "sub": 1}),
        ("DPT-1 DPT-5", {"main": 1, "sub": None}),
        ("DPT-14 DPST-14-1", {"main": 14, "sub": None}),
        ("DPST-6-10", {"main": 6, "sub": 10}),
        ("Wrong", None),
        ("DPT-Wrong", None),
        ("DPST-1-Wrong", None),
        ("DPST-5", None),
        ([], None),
        (None, None),
    ],
)
def test_get_dpt_type(
    dpt_string: str | list[Any] | None, expected: dict[str, int | None] | None
) -> None:
    """Test parsing single DPT from ETS project."""
    assert util.get_dpt_type(dpt_string) == expected


@pytest.mark.parametrize(
    ("dpt_string", "expected"),
    [
        ("DPT-1", [{"main": 1, "sub": None}]),
        ("DPT-1 DPST-1-1", [{"main": 1, "sub": None}, {"main": 1, "sub": 1}]),
        ("DPT-7 DPST-7-1", [{"main": 7, "sub": None}, {"main": 7, "sub": 1}]),
        ("DPST-5-1", [{"main": 5, "sub": 1}]),
        ("DPT-1 DPT-5", [{"main": 1, "sub": None}, {"main": 5, "sub": None}]),
        ("DPT-14 DPST-14-1", [{"main": 14, "sub": None}, {"main": 14, "sub": 1}]),
        ("DPST-6-10", [{"main": 6, "sub": 10}]),
        ("Wrong", []),
        ("DPT-Wrong", []),
        ("DPST-1-Wrong", []),
        ("DPST-5", []),
        ([], []),
        (None, []),
    ],
)
def test_parse_dpt_types(
    dpt_string: str | list[Any] | None, expected: list[DPTType]
) -> None:
    """Test parsing list of DPT from ETS project."""
    assert util.parse_dpt_types(dpt_string) == expected


@pytest.mark.parametrize(
    ("text", "parameter", "expected"),
    [
        ("{{0}}", ParameterInstanceRef("id", "test"), "test"),
        ("{{0:default}}", ParameterInstanceRef("id", None), "default"),
        ("{{0:default}}", ParameterInstanceRef("id", "test"), "test"),
        ("{{0}}", None, ""),
        ("{{0:default}}", None, "default"),
        ("Hello {{0}}", ParameterInstanceRef("id", "test"), "Hello test"),
        ("Hi {{0:def}} again", ParameterInstanceRef("id", None), "Hi def again"),
        ("Hi{{0:default}}again", ParameterInstanceRef("id", "test"), "Hitestagain"),
        ("{{1}}", ParameterInstanceRef("id", "test"), "{{1}}"),
        ("{{XY}}:{{0}}{{ZZ}}", ParameterInstanceRef("id", "test"), "{{XY}}:test{{ZZ}}"),
    ],
)
def test_text_parameter_template_replace(
    text: str, parameter: ParameterInstanceRef | None, expected: str
) -> None:
    """Test strip_module_instance."""
    assert util.text_parameter_template_replace(text, parameter) == expected


@pytest.mark.parametrize(
    ("text", "search_id", "expected"),
    [
        ("CH-4", "CH", "CH-4"),
        ("MD-1_M-1_MI-1_CH-4", "CH", "MD-1_CH-4"),
        ("MD-4_M-15_MI-1_SM-1_M-1_MI-1-1-2_SM-1_O-3-1_R-2", "O", "MD-4_SM-1_O-3-1_R-2"),
    ],
)
def test_strip_module_instance(text: str, search_id: str, expected: str) -> None:
    """Test strip_module_instance."""
    assert util.strip_module_instance(text, search_id) == expected


@pytest.mark.parametrize(
    ("ref", "next_id", "expected"),
    [
        ("M-0083_A-0098-12-489B_MD-1_M-1_MI-1_P-43_R-87", "P", "MD-1_M-1_MI-1"),
        ("MD-1_M-1_MI-1_CH-4", "CH", "MD-1_M-1_MI-1"),
        (
            "MD-4_M-15_MI-1_SM-1_M-1_MI-1-1-2_SM-1_O-3-1_R-2",
            "O",
            "MD-4_M-15_MI-1_SM-1_M-1_MI-1-1-2_SM-1",
        ),
        ("M-00FA_A-A228-0A-A6C3_O-2002002_R-200200202", "O", ""),
        ("MD-1_M-1_MI-1_CH-4", "CH", "MD-1_M-1_MI-1"),
        ("CH-SOM03", "CH", ""),
    ],
)
def test_get_module_instance_part(ref: str, next_id: str, expected: str) -> None:
    """Test strip_module_instance."""
    assert util.get_module_instance_part(ref, next_id) == expected


@pytest.mark.parametrize(
    ("instance_ref", "instance_next_id", "text_parameter_ref_id", "expected"),
    [
        (
            "MD-2_M-17_MI-1_O-3-0_R-159",
            "O",
            "M-0083_A-00B0-32-0DFC_MD-2_P-23_R-1",
            "M-0083_A-00B0-32-0DFC_MD-2_M-17_MI-1_P-23_R-1",
        ),
        (
            "MD-2_M-6_MI-1_CH-1",
            "CH",
            "M-0083_A-013A-32-DCC1_MD-2_P-1_R-1",
            "M-0083_A-013A-32-DCC1_MD-2_M-6_MI-1_P-1_R-1",
        ),
        (
            "O-595_R-688",
            "O",
            "M-0004_A-20D3-11-EC49-O000A_P-875_R-2697",  # no module - return same string,
            "M-0004_A-20D3-11-EC49-O000A_P-875_R-2697",
        ),
        (
            "MD-5_M-2_MI-1_O-3-0_R-1",
            "O",
            "M-007C_A-0004-72-F374_MD-5_UP-3_R-3",  # UnionParameter
            "M-007C_A-0004-72-F374_MD-5_M-2_MI-1_UP-3_R-3",
        ),
    ],
)
def test_text_parameter_insert_module_instance(
    instance_ref: str, instance_next_id: str, text_parameter_ref_id: str, expected: str
) -> None:
    """Test strip_module_instance."""
    assert (
        util.text_parameter_insert_module_instance(
            instance_ref, instance_next_id, text_parameter_ref_id
        )
        == expected
    )



"""XML utilities."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, overload

from xknxproject.const import MAIN_AND_SUB_DPT, MAIN_DPT
from xknxproject.exceptions import UnexpectedDataError
from xknxproject.models import DPTType

if TYPE_CHECKING:
    from xknxproject.models import ParameterInstanceRef

_LOGGER = logging.getLogger("xknxproject.log")


def get_dpt_type(dpt_string: str | None) -> DPTType | None:
    """Parse DPT type from the XML representation to main and sub types."""
    # GroupAddress tags should only support one single DPT.
    try:
        return parse_dpt_types(dpt_string)[0]
    except IndexError:
        return None


def parse_dpt_types(dpt_string: str | None) -> list[DPTType]:
    """Parse all DPTs from the XML representation to main and sub types."""
    if not dpt_string:
        return []

    supported_dpts: list[DPTType] = []
    # some applications have listed same DPT multiple times `DatapointType="DPST-1-1 DPST-1-1"`
    # so we use dict.fromkeys() (as set() doesn't preserve order)
    for _dpt in dict.fromkeys(dpt_string.split()):
        dpt_parts = _dpt.split("-")
        try:
            if dpt_parts[0] == MAIN_DPT:
                supported_dpts.append(
                    DPTType(
                        main=int(dpt_parts[1]),
                        sub=None,
                    )
                )
            if dpt_parts[0] == MAIN_AND_SUB_DPT:
                supported_dpts.append(
                    DPTType(
                        main=int(dpt_parts[1]),
                        sub=int(dpt_parts[2]),
                    )
                )
        except (IndexError, ValueError):
            _LOGGER.warning(
                'Could not parse DPTType from: "%s" in "%s"', _dpt, dpt_string
            )
    return supported_dpts


@overload
def parse_xml_flag(flag: str | None, default: bool) -> bool: ...


@overload
def parse_xml_flag(flag: str | None, default: None = None) -> bool | None: ...


def parse_xml_flag(flag: str | None, default: bool | None = None) -> bool | None:
    """Parse the XML flag to an optional boolean."""
    if flag is None:
        return default
    return flag == "Enabled"


def text_parameter_template_replace(
    text: str, parameter: ParameterInstanceRef | None
) -> str:
    """Replace parameter template in text."""
    # Text of a Channel, ParameterBlock, ParameterSeparator, ParameterRef or ComObjectRef
    # may use placeholder "{{0}}" or "{{0:def}}" (without the quotes). def is a default
    # text to be displayed if the text parameter value is empty.
    # These placeholders (with or without the default text) are included in translations too.

    # Applications TextParameterRef points to 0.xml ParameterInstanceRef of DeviceInstance

    parameter_value = parameter.value if parameter is not None else None
    return re.sub(
        r"{{0(?::?)(.*?)}}",
        lambda matchobj: parameter_value or matchobj.group(1),
        text,
    )


def strip_module_instance(text: str, search_id: str) -> str:
    """
    Remove module and module instance from text, keep module definition and rest.

    text: full text to be processed
    search_id: search term to be kept without "-" eg. "CH" for channel

    Examples
    --------
    search_id="CH": "CH-4" -> "CH-4"
    search_id="CH": "MD-1_M-1_MI-1_CH-4" -> "MD-1_CH-4"
    search_id="O": "MD-4_M-15_MI-1_SM-1_M-1_MI-1-1-2_SM-1_O-3-1_R-2" -> "MD-4_SM-1_O-3-1_R-2"

    """
    # For submodules SM- must be the last item before search_id
    # because I couldn't create a regex that works otherwise :(
    return re.sub(
        r"(MD-\w+_)?.*?(SM-\w+_)?(" + re.escape(search_id) + r"-.*)",
        lambda matchobj: "".join(part for part in matchobj.groups() if part),
        text,
    )


def get_module_instance_part(ref: str, next_id: str) -> str:
    """
    Get module and module instance from text or empty string if not found.

    ref: full text to be processed
    next_id: search term after module definitions. Eg. "CH" for channel

    """
    # For submodules SM- must be the last item before search_id
    # because I couldn't create a regex that works otherwise :(

    matchobj = re.search(r"(MD-.*)_" + re.escape(next_id) + r"-", ref)
    return matchobj.group(1) if matchobj else ""


def text_parameter_insert_module_instance(
    instance_ref: str, instance_next_id: str, text_parameter_ref_id: str
) -> str:
    """
    Insert module and module instance from instance_ref into target_ref.

    instance_ref: reference holding module instance
    instance_next_id: search term after module definitions. Eg. "CH" for channel
    text_parameter_ref_id: reference with module definition where module instance
      should be inserted after module definition
    """
    if "_MD-" in text_parameter_ref_id and (
        _module_ref := get_module_instance_part(instance_ref, next_id=instance_next_id)
    ):
        _application_ref = text_parameter_ref_id.split("_MD-", maxsplit=1)[0]
        try:
            # `_P-` for Parameter `_UP-` for UnionParameter
            _parameter_ref = re.search(r"_(U?P-.*)", text_parameter_ref_id).group(1)  # type: ignore[union-attr]
        except AttributeError:
            raise UnexpectedDataError(
                f"No Parameter block found in TextParameterRefId {text_parameter_ref_id} "
                f"(instance: {instance_ref})"
            ) from None
        return f"{_application_ref}_{_module_ref}_{_parameter_ref}"

    return text_parameter_ref_id



"""Define internally used data structures."""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field
import logging
import re

from xknxproject import util
from xknxproject.exceptions import UnexpectedDataError
from xknxproject.models.knxproject import DPTType, ModuleInstanceInfos
from xknxproject.models.static import GroupAddressStyle, SpaceType
from xknxproject.zip import KNXProjContents

_LOGGER = logging.getLogger("xknxproject.log")

TranslationsType = dict[str, dict[str, str]]


class XMLGroupAddress:
    """Class that represents a group address."""

    def __init__(
        self,
        name: str,
        identifier: str,
        address: str,
        project_uid: int | None,
        description: str,
        dpt: DPTType | None,
        data_secure_key: str | None,
        comment: str,
        style: GroupAddressStyle,
    ) -> None:
        """Initialize a group address."""
        self.name = name
        self.identifier = identifier.split("_", 1)[1]
        self.raw_address = int(address)
        self.project_uid = project_uid
        self.description = description
        self.dpt = dpt
        self.data_secure_key = data_secure_key  # Key as base64 encoded string or None
        self.comment = comment
        self.style = style
        self.address = XMLGroupAddress.str_address(self.raw_address, self.style)

    @staticmethod
    def str_address(raw_address: int, group_address_style: GroupAddressStyle) -> str:
        """Parse a given address and returns a string representation of it."""
        if group_address_style == GroupAddressStyle.FREE:
            return str(raw_address)
        main = (raw_address & 0b1111100000000000) >> 11
        if group_address_style == GroupAddressStyle.THREELEVEL:
            middle = (raw_address & 0b11100000000) >> 8
            sub = raw_address & 0b11111111
            return f"{main}/{middle}/{sub}"
        if group_address_style == GroupAddressStyle.TWOLEVEL:
            sub = raw_address & 0b11111111111
            return f"{main}/{sub}"
        raise ValueError(f"GroupAddressSyste '{group_address_style}' not supported!")

    def __repr__(self) -> str:
        """Return string representation."""
        return (
            f"{self.address} ({self.name}) - [DPT: {self.dpt}, ID: {self.identifier}]"
        )


@dataclass
class XMLGroupRange:
    """Class that represents a group range."""

    name: str
    range_start: int
    range_end: int
    group_addresses: list[int]
    group_ranges: list[XMLGroupRange]
    comment: str
    style: GroupAddressStyle

    def str_address(self) -> str:
        """Generate a string representation for the range."""
        if self.style == GroupAddressStyle.FREE:
            return f"{self.range_start}...{self.range_end}"
        if self.style == GroupAddressStyle.TWOLEVEL:
            return XMLGroupAddress.str_address(self.range_start, self.style).split(
                "/", maxsplit=1
            )[0]
        if self.style == GroupAddressStyle.THREELEVEL:
            start_address_token = XMLGroupAddress.str_address(
                self.range_start, self.style
            ).split("/")
            if (self.range_end - self.range_start) >= 2046:
                return start_address_token[0]
            return "/".join(start_address_token[0:2])
        raise ValueError(f"GroupAddressSyste '{self.style}' not supported!")


@dataclass
class XMLArea:
    """Class that represents a area."""

    address: int
    name: str
    description: str | None
    lines: list[XMLLine]


@dataclass
class XMLLine:
    """Class that represents a Line."""

    address: int
    description: str | None
    name: str
    medium_type: str
    devices: list[DeviceInstance]
    area: XMLArea


class DeviceInstance:
    """Class that represents a device instance."""

    def __init__(
        self,
        *,
        identifier: str,
        address: int,
        project_uid: int | None,
        name: str,
        description: str,
        last_modified: str,
        product_ref: str,
        hardware_program_ref: str,
        line: XMLLine,
        manufacturer: str,
        additional_addresses: list[str],
        channels: list[ChannelNode],
        com_object_instance_refs: list[ComObjectInstanceRef],
        module_instances: list[ModuleInstance],
        parameter_instance_refs: dict[str, ParameterInstanceRef],
        com_objects: list[ComObject] | None = None,
    ) -> None:
        """Initialize a Device Instance."""
        self.identifier = identifier
        self.address = address
        self.name = name  # empty string if not customized in project
        self.description = description
        self.project_uid = project_uid
        self.last_modified = last_modified
        self.product_ref = product_ref
        self.hardware_program_ref = hardware_program_ref
        self.line = line
        self.area_address = line.area.address  # used for sorting
        self.line_address = line.address  # used for sorting
        self.manufacturer = manufacturer
        self.additional_addresses = additional_addresses
        self.channels: list[ChannelNode] = channels
        self.com_object_instance_refs = com_object_instance_refs
        self.module_instances = module_instances
        self.com_objects = com_objects or []
        self.parameter_instance_refs = parameter_instance_refs
        self.application_program_ref: str | None = None

        self.individual_address = (
            f"{self.area_address}.{self.line_address}.{self.address}"
        )
        self.product_name: str = ""  # translatable name for specific product
        self.hardware_name: str = ""  # untranslatable name from hardware.xml
        self.order_number: str = ""
        self.manufacturer_name: str = ""

    def add_additional_address(self, address: str) -> None:
        """Add an additional individual address."""
        self.additional_addresses.append(
            f"{self.line.area.address}/{self.line.address}/{address}"
        )

    def application_program_xml(self) -> str:
        """Obtain the file name to the application program XML."""
        return f"{self.manufacturer}/{self.application_program_ref}.xml"

    def module_instance_arguments(self) -> Iterator[ModuleInstanceArgument]:
        """Iterate ModuleInstance arguments."""
        for _module_instance in self.module_instances:
            yield from _module_instance.arguments

    def merge_application_program_info(self, application: ApplicationProgram) -> None:
        """Merge items with their parent objects from the application program."""
        for argument in self.module_instance_arguments():
            argument.name = application.module_def_arguments[argument.ref_id].name
            argument.allocates = application.module_def_arguments[
                argument.ref_id
            ].allocates

        for com_instance in self.com_object_instance_refs:
            com_instance.merge_application_program_info(
                application, self.parameter_instance_refs
            )
            com_instance.apply_module_base_number_argument(
                module_instances=self.module_instances,
                application=application,
            )

        for channel in self.channels:
            channel.resolve_channel_name(device_instance=self, application=application)
            channel.resolve_channel_module_placeholders(device_instance=self)

    def __str__(self) -> str:
        """Return string representation."""
        return (
            "DeviceInstance("
            f'"{self.individual_address} {self.manufacturer_name} - {self.product_name}" '
            f'Id="{self.identifier}" Puid="{self.project_uid}" '
            f'ApplicationProgram="{self.application_program_ref}")'
        )


@dataclass
class ChannelNode:
    """Class that represents a Node with Type Channel."""

    ref_id: str  # name="RefId" type="knx:RELIDREF" use="required"
    name: str  # name="Text" type="xs:string" use="optional"
    group_object_instances: list[
        str
    ]  # name="GroupObjectInstances" type="knx:RELIDREFS" use="optional"

    def resolve_channel_name(
        self,
        device_instance: DeviceInstance,
        application: ApplicationProgram,
    ) -> None:
        """
        Resolve the channel name from device instance infos.

        Replace TextParameter values in channel names with the
        actual values of the parameter instances.
        """
        if not self.name:
            application_channel_id = util.strip_module_instance(
                self.ref_id, search_id="CH"
            )
            application_channel = application.channels[
                f"{device_instance.application_program_ref}_{application_channel_id}"
            ]
            if application_channel.text and application_channel.text_parameter_ref_id:
                parameter_instance_ref = util.text_parameter_insert_module_instance(
                    instance_ref=self.ref_id,
                    instance_next_id="CH",
                    text_parameter_ref_id=application_channel.text_parameter_ref_id,
                )
                try:
                    parameter = device_instance.parameter_instance_refs[
                        parameter_instance_ref
                    ]
                except KeyError:
                    _LOGGER.debug(
                        "ParameterInstanceRef %s not found for Channel %s of %s",
                        parameter_instance_ref,
                        self.ref_id,
                        device_instance,
                    )
                    parameter = None

                self.name = (
                    util.text_parameter_template_replace(
                        application_channel.text, parameter
                    )
                    or application_channel.name
                )
            else:
                self.name = application_channel.text or application_channel.name

    def resolve_channel_module_placeholders(
        self,
        device_instance: DeviceInstance,
    ) -> None:
        """Replace module placeholders in channel names with module instance argument values."""
        if not (
            self.ref_id.startswith("MD-")  # only applicable if modules used
            and "{{" in self.name  # placeholders are denoted "{{name}}"
        ):
            return

        module_instance_ref = self.ref_id.split("_CH", maxsplit=1)[0]
        try:
            module_instance = next(
                mi
                for mi in device_instance.module_instances
                if mi.identifier == module_instance_ref
            )
        except StopIteration:
            raise UnexpectedDataError(
                f"ModuleInstance '{module_instance_ref}' not found for "
                f"ChannelNode '{self.ref_id}' {self.name} of {device_instance}"
            ) from None
        for argument in module_instance.arguments:
            self.name = self.name.replace(f"{{{{{argument.name}}}}}", argument.value)


@dataclass
class ModuleInstance:
    """Class that represents a ModuleInstance."""

    identifier: str  # MD-4_M-15_MI-2 / MD-4_M-15_MI-2_SM-1_M-1_MI-2-1-2
    ref_id: str  # MD-<int>_M-<int>
    arguments: list[ModuleInstanceArgument]

    module_def_id: str = field(init=False)  # MD-<int>
    # MD-<int>_M-<int>_MI-<int> if SubModule - reference may contain base values for arguments
    base_module: str | None = field(init=False)
    # ModuleDefUniqueNumber with SubModule id if present (used in result json)
    definition_id: str = field(init=False)  # MD-<int>[_SM-<int>]

    def __post_init__(self) -> None:
        """Set is_submodule based on the identifier."""
        self.module_def_id = self.ref_id.split("_", maxsplit=1)[0]
        _submodule_match = re.search(r"(_SM-[^_]+)", self.identifier)
        if _submodule_match is not None:
            self.base_module = f"{self.identifier.split('_SM-', maxsplit=1)[0]}"
            self.definition_id = f"{self.module_def_id}{_submodule_match.group()}"
        else:
            self.base_module = None
            self.definition_id = self.module_def_id

    def complete_arguments_ref_id(self, application_program_ref: str) -> None:
        """Prepend the ref_id with the application program ref."""
        for arg in self.arguments:
            arg.complete_ref_id(application_program_ref, self.module_def_id)


@dataclass
class ModuleInstanceArgument:
    """Class that represents a ModuleInstance Argument."""

    ref_id: str
    value: str
    # resolved from application by `ref_id` ModuleDefs/ModuleDef/Arguments/Argument
    name: str = ""  # "Name" type="knx:Identifier50_t" use="required"
    allocates: int | None = None  # "Allocates" type="xs:unsignedLong" use="optional"

    def complete_ref_id(self, application_program_ref: str, module_def_id: str) -> None:
        """Prepend the ref_id with the application program ref."""
        if self.ref_id.startswith("SM-"):  # SubModule
            self.ref_id = f"{application_program_ref}_{module_def_id}_{self.ref_id}"
        else:
            self.ref_id = f"{application_program_ref}_{self.ref_id}"


@dataclass
class ComObjectInstanceRef:
    """Class that represents a ComObjectInstanceRef instance."""

    identifier: str | None  # "Id" - xs:ID
    # "RefId" - knx:RELIDREF - required - points to a ComObjectRef Id
    # initially stripped by the devices application_program_ref
    ref_id: str
    text: str | None  # "Text"
    function_text: str | None  # "FunctionText"
    read_flag: bool | None  # "ReadFlag" - knx:Enable_t
    write_flag: bool | None  # "WriteFlag" - knx:Enable_t
    communication_flag: bool | None  # "CommunicationFlag" - knx:Enable_t
    transmit_flag: bool | None  # "TransmitFlag" - knx:Enable_t
    update_flag: bool | None  # "UpdateFlag" - knx:Enable_t
    read_on_init_flag: bool | None  # "ReadOnInitFlag" - knx:Enable_t
    datapoint_types: list[DPTType]  # "DataPointType" - knx:IDREFS
    description: str | None  # "Description" - language dependent
    channel: str | None  # "ChannelId" - knx:IDREFS
    links: list[str] | None  # "Links" - knx:RELIDREFS

    # resolved via Hardware.xml from the containing DeviceInstance
    application_program_id_prefix: str = (
        ""  # empty string for ETS 4 since it doesn't use prefix
    )
    com_object_ref_id: str | None = None

    # only available form ComObject and ComObjectRef
    name: str | None = None
    object_size: str | None = None
    # only available form ComObject
    base_number_argument_ref: str | None = None  # optional in ComObject
    number: int | None = None  # required in ComObject
    # assigned when module arguments are applied
    module: ModuleInstanceInfos | None = None

    def resolve_com_object_ref_id(
        self, application_program_ref: str, knx_proj_contents: KNXProjContents
    ) -> None:
        """Prepend the ref_id with the application program ref."""
        if knx_proj_contents.is_ets4_project() and self.ref_id.startswith(
            application_program_ref
        ):
            # ETS4 doesn't use shortened ref_id and I think it doesn't support modules at all
            self.com_object_ref_id = self.ref_id
            return

        ref_id = util.strip_module_instance(self.ref_id, search_id="O")
        self.application_program_id_prefix = f"{application_program_ref}_"
        self.com_object_ref_id = f"{application_program_ref}_{ref_id}"

    def merge_application_program_info(
        self,
        application: ApplicationProgram,
        parameters: dict[str, ParameterInstanceRef],
    ) -> None:
        """Fill missing information with information parsed from the application program."""
        if self.com_object_ref_id is None:
            _LOGGER.warning(
                "ComObjectInstanceRef %s has no ComObjectRefId",
                self.identifier,
            )
            return
        com_object_ref = application.com_object_refs[self.com_object_ref_id]
        self._merge_from_parent_object(com_object_ref, parameters=parameters)

        com_object = application.com_objects[com_object_ref.ref_id]
        self._merge_from_parent_object(com_object, parameters=parameters)

    def _merge_from_parent_object(
        self,
        com_object: ComObject | ComObjectRef,
        parameters: dict[str, ParameterInstanceRef],
    ) -> None:
        """Fill missing information with information parsed from the application program."""
        if self.name is None:
            self.name = com_object.name
        if self.text is None:
            if isinstance(com_object, ComObjectRef):
                self.text = (
                    com_object.com_object_ref_text_with_paramter(
                        com_object_instance_ref_id=self.ref_id,
                        instance_parameters=parameters,
                    )
                    or com_object.text
                )
            else:
                self.text = com_object.text
        if self.function_text is None:
            self.function_text = com_object.function_text
        if self.object_size is None:
            self.object_size = com_object.object_size
        if self.read_flag is None:
            self.read_flag = com_object.read_flag
        if self.write_flag is None:
            self.write_flag = com_object.write_flag
        if self.communication_flag is None:
            self.communication_flag = com_object.communication_flag
        if self.transmit_flag is None:
            self.transmit_flag = com_object.transmit_flag
        if self.update_flag is None:
            self.update_flag = com_object.update_flag
        if self.read_on_init_flag is None:
            self.read_on_init_flag = com_object.read_on_init_flag
        if not self.datapoint_types:
            self.datapoint_types = com_object.datapoint_types
        if isinstance(com_object, ComObject):
            self.number = com_object.number
            self.base_number_argument_ref = com_object.base_number_argument_ref

    def apply_module_base_number_argument(
        self,
        module_instances: list[ModuleInstance],
        application: ApplicationProgram,
    ) -> None:
        """Apply module argument of base number."""
        if (
            self.base_number_argument_ref is None
            or not self.ref_id.startswith("MD-")
            or self.number is None  # only for type safety
        ):
            return

        def _parse_base_number_argument(
            module_instance: ModuleInstance,
            base_number_argument_ref: str,
        ) -> int:
            """Parse the argument value."""
            # there are 2 ways to get the base number
            # 1. from the module instance arguments value "ObjNumberBase" directly
            # 2. from the module defs allocator - adding the "Start" value to
            #   (the modules index - 1) * allocator size
            #   in this case the module instance argument value is the reference part
            #   of the allocator id ("L-1") - at least in the application tested (MDT Dali 64)
            #
            #   for SubModules the NumericArg item may use a BaseValue reference to an
            #   Argument of the base ModuleDef containing the base value for all its SubModules
            result = 0
            try:
                base_number_argument = next(
                    arg
                    for arg in module_instance.arguments
                    if arg.ref_id == base_number_argument_ref
                )
            except StopIteration:
                raise UnexpectedDataError(
                    f"Base number argument {base_number_argument_ref} not found for "
                    f"ComObjectInstanceRef {self.ref_id=} {self.text=} "
                    f"of application {self.application_program_id_prefix}",
                ) from None

            try:
                # path (1) if value is a number, we are done
                # base module value should already be included
                return int(base_number_argument.value)
            except ValueError:
                # path (2) value is a reference to an Allocator
                if module_instance.base_module:
                    # recurse to get the base number from the base module (for SubModule value)
                    num_arg = application.numeric_args.get(base_number_argument.ref_id)
                    if (
                        num_arg is not None
                        and (base_value_ref := num_arg.base_value) is not None
                    ):
                        try:
                            base_module = next(
                                mi
                                for mi in module_instances
                                if mi.identifier == module_instance.base_module
                            )
                        except StopIteration:
                            raise UnexpectedDataError(
                                f"Base ModuleInstance {module_instance.base_module} not found for "
                                f"ComObjectInstanceRef {self.ref_id=} {self.text=} "
                                f"of application {self.application_program_id_prefix}",
                            ) from None
                        result += _parse_base_number_argument(
                            module_instance=base_module,
                            base_number_argument_ref=base_value_ref,
                        )
                return result + self._base_number_from_allocator(
                    base_number_argument, application.allocators
                )

        try:
            _module_instance = next(
                mi
                for mi in module_instances
                if self.ref_id.startswith(f"{mi.identifier}_")
            )
        except StopIteration:
            raise UnexpectedDataError(
                f"ModuleInstance not found for ComObjectInstanceRef {self.ref_id=} {self.text=} "
                f"of application {self.application_program_id_prefix}",
            ) from None

        com_object_number = self.number
        self.number += _parse_base_number_argument(
            _module_instance, self.base_number_argument_ref
        )
        self.module = ModuleInstanceInfos(
            definition=_module_instance.definition_id,
            root_number=com_object_number,
        )

    def _base_number_from_allocator(
        self,
        base_number_argument: ModuleInstanceArgument,
        application_allocators: dict[str, Allocator],
    ) -> int:
        """Apply base number from allocator."""
        try:
            allocator_object_base = next(
                allocator
                for allocator in application_allocators.values()
                if allocator.identifier
                == self.application_program_id_prefix + base_number_argument.value
            )
        except StopIteration:
            raise UnexpectedDataError(
                f"Allocator with identifier {base_number_argument.value} not found for "
                f"ComObjectInstanceRef {self.ref_id=} {self.text=} "
                f"of application {self.application_program_id_prefix}",
            ) from None
        if (allocator_size := base_number_argument.allocates) is None:
            _LOGGER.warning(
                "Base number allocator size not found for %s. Base number argument: %s",
                self.identifier,
                base_number_argument,
            )
            return 0
        module_instance_index = int(
            self.ref_id.split("_MI-", maxsplit=1)[1].split("_", maxsplit=1)[0]
        )
        return allocator_object_base.start + (
            allocator_size * (module_instance_index - 1)
        )


@dataclass
class ParameterInstanceRef:
    """ParameterInstanceRef."""

    ref_id: str
    value: str | None


@dataclass
class ApplicationProgram:
    """Class that represents an ApplicationProgram instance."""

    com_objects: dict[str, ComObject]  # {Id: ComObject}
    com_object_refs: dict[str, ComObjectRef]  # {Id: ComObjectRef}
    allocators: dict[str, Allocator]  # {Id: Allocator}
    module_def_arguments: dict[str, ModuleDefinitionArgumentInfo]  # {Id: ...}
    numeric_args: dict[str, ModuleDefinitionNumericArg]  # {RefId: ...}
    channels: dict[str, ApplicationProgramChannel]  # {Id: ApplicationProgramChannel}


@dataclass
class Allocator:
    """Allocator."""

    identifier: str
    name: str
    start: int
    end: int


@dataclass
class ModuleDefinitionArgumentInfo:
    """Module Definition Argument."""

    name: str = ""
    allocates: int | None = None


@dataclass
class ModuleDefinitionNumericArg:
    """Module Definition Numeric Argument."""

    # shortened version (MD-<int>_L-<int>) should be Value in 0.xml ModuleInstanceArgument (at least with ETS 5)
    allocator_ref_id: str | None
    # if allocator_ref_id is not used, this is 0.xml ModuleInstanceArgument Value
    value: int | None
    # RefId to Argument (<application>_MD-<int>_A-<int>) - Base value for arguments used in SubModules
    base_value: str | None


@dataclass
class ApplicationProgramChannel:
    """ApplicationProgramChannel."""

    __slots__ = (
        "identifier",
        "name",
        "number",
        "text",
        "text_parameter_ref_id",
    )

    identifier: str  # name="Id" type="xs:ID" use="required"
    text: (
        str | None
    )  # name="Text" type="knx:LanguageDependentString255_t" use="optional"
    text_parameter_ref_id: (
        str | None
    )  # name="TextParameterRefId" type="knx:RELIDREF" use="optional"
    name: str  # name="Name" type="knx:String255_t" use="required"
    number: str  # name="Number" type="knx:String50_t" use="required"


@dataclass
class ComObject:
    """Class that represents a ComObject instance."""

    __slots__ = (
        "base_number_argument_ref",
        "communication_flag",
        "datapoint_types",
        "function_text",
        "identifier",
        "name",
        "number",
        "object_size",
        "read_flag",
        "read_on_init_flag",
        "text",
        "transmit_flag",
        "update_flag",
        "write_flag",
    )

    # all items required in the XML
    identifier: str  # "Id" - xs:ID
    name: str  # "Name"
    text: str  # "Text" - language dependent
    number: int  # "Number" - xs:unsignedInt
    function_text: str  # "FunctionText" - language dependent
    object_size: str  # "ObjectSize" - knx:ComObjectSize_t
    read_flag: bool  # "ReadFlag" - knx:Enable_t
    write_flag: bool  # "WriteFlag" - knx:Enable_t
    communication_flag: bool  # "CommunicationFlag" - knx:Enable_t
    transmit_flag: bool  # "TransmitFlag" - knx:Enable_t
    update_flag: bool  # "UpdateFlag" - knx:Enable_t
    read_on_init_flag: bool  # "ReadOnInitFlag" - knx:Enable_t
    datapoint_types: list[DPTType]  # "DataPointType" - knx:IDREFS - optional
    # "BaseNumber" - knx:IDREF - optional - schema version >= 20
    # ModuleArgument identifier that holds value to add for
    # communication object number of ComObjectInstanceRef
    base_number_argument_ref: str | None


@dataclass
class ComObjectRef:
    """Class that represents a ComObjectRef instance."""

    __slots__ = (
        "communication_flag",
        "datapoint_types",
        "function_text",
        "identifier",
        "name",
        "object_size",
        "read_flag",
        "read_on_init_flag",
        "ref_id",
        "text",
        "text_parameter_ref_id",
        "transmit_flag",
        "update_flag",
        "write_flag",
    )

    identifier: str  # "Id" - xs:ID - required
    ref_id: str  # "RefId" - knx:IDREF - required - points to a ComObject Id
    name: str | None  # "Name"
    text: str | None  # "Text" - language dependent
    function_text: str | None  # "FunctionText" - language dependent
    object_size: str | None  # "ObjectSize" - knx:ComObjectSize_t
    read_flag: bool | None  # "ReadFlag" - knx:Enable_t
    write_flag: bool | None  # "WriteFlag" - knx:Enable_t
    communication_flag: bool | None  # "CommunicationFlag" - knx:Enable_t
    transmit_flag: bool | None  # "TransmitFlag" - knx:Enable_t
    update_flag: bool | None  # "UpdateFlag" - knx:Enable_t
    read_on_init_flag: bool | None  # "ReadOnInitFlag" - knx:Enable_t
    datapoint_types: list[DPTType]  # "DataPointType" - knx:IDREFS
    text_parameter_ref_id: str | None  #  type="knx:IDREF" use="optional"

    def com_object_ref_text_with_paramter(
        self,
        com_object_instance_ref_id: str,
        instance_parameters: dict[str, ParameterInstanceRef],
    ) -> str | None:
        """Return the text with parameter if available."""
        if self.text and self.text_parameter_ref_id:
            parameter_instance_ref = util.text_parameter_insert_module_instance(
                instance_ref=com_object_instance_ref_id,
                instance_next_id="O",
                text_parameter_ref_id=self.text_parameter_ref_id,
            )
            try:
                parameter = instance_parameters[parameter_instance_ref]
            except KeyError:
                _LOGGER.debug(
                    "ParameterInstanceRef %s for ComObjectRef %s not found.",
                    parameter_instance_ref,
                    self.identifier,
                )
                parameter = None
            return util.text_parameter_template_replace(
                self.text or "",
                parameter=parameter,
            )
        return None


@dataclass
class KNXMasterData:
    """KNX Master data needed for parsing other project files."""

    function_type_names: dict[str, str]
    manufacturer_names: dict[str, str]
    space_usage_mapping: dict[str, str]
    translations: TranslationsType

    def _get_translation_item(
        self, ref_id: str, attribute_name: str = "Text"
    ) -> str | None:
        """Get translation item from the translations dict."""
        if self.translations:
            try:
                return self.translations[ref_id][attribute_name]
            except KeyError:
                return None
        return None

    def get_function_type_name(self, function_type_id: str) -> str:
        """Get space usage name from space usage id."""
        if translated := self._get_translation_item(function_type_id):
            return translated
        return self.function_type_names.get(function_type_id, "")

    def get_space_usage_name(self, space_usage_id: str) -> str:
        """Get space usage name from space usage id."""
        if translated := self._get_translation_item(space_usage_id):
            return translated
        return self.space_usage_mapping.get(space_usage_id, "")


@dataclass
class XMLSpace:
    """A space in the location XML."""

    identifier: str
    name: str
    space_type: SpaceType
    usage_id: str | None  # SU-<int> resolved from knx_master.xml (with translation)
    usage_text: str  # default to "" - translated
    number: str  # default to ""
    description: str  # default to ""
    project_uid: int | None
    spaces: list[XMLSpace]
    devices: list[str]  # [DeviceInstance.individual_address]
    functions: list[str]


@dataclass
class XMLFunction:
    """A functions in the space XML."""

    function_type: str
    group_addresses: list[XMLGroupAddressRef]
    identifier: str
    name: str
    project_uid: int | None
    space_id: str
    usage_text: str


@dataclass
class XMLGroupAddressRef:
    """A GroupAddressRef in the functions XML."""

    address: str
    identifier: str
    name: str
    project_uid: int | None
    ref_id: str
    role: str


@dataclass
class Product:
    """Model a Product instance."""

    identifier: str
    text: str
    order_number: str
    hardware_name: str = ""


HardwareToPrograms = dict[str, str]


@dataclass
class XMLProjectInformation:
    """Model a ProjectInformation instance."""

    # ProjectInformation tag is not required in XSD, thus everything is optional
    project_id: str = ""
    name: str = ""
    last_modified: str | None = None
    group_address_style: GroupAddressStyle = GroupAddressStyle.THREELEVEL
    guid: str = ""
    created_by: str = ""
    schema_version: str = ""
    tool_version: str = ""



"""Application Program Loader."""

from __future__ import annotations

from collections.abc import Iterator
from typing import Any
from xml.etree import ElementTree
from zipfile import Path

from xknxproject.models import (
    Allocator,
    ApplicationProgram,
    ApplicationProgramChannel,
    ComObject,
    ComObjectRef,
    DeviceInstance,
    ModuleDefinitionArgumentInfo,
    ModuleDefinitionNumericArg,
)
from xknxproject.util import parse_dpt_types, parse_xml_flag


class ApplicationProgramLoader:
    """Load the application program from KNX XML."""

    @staticmethod
    def load(
        application_program_path: Path,
        devices: list[DeviceInstance],
        language_code: str | None,
    ) -> ApplicationProgram:
        """Load Hardware mappings and assign to devices."""
        com_object_instance_refs = [
            instance_ref
            for device in devices
            for instance_ref in device.com_object_instance_refs
        ]
        used_com_object_ref_ids = {
            instance_ref.com_object_ref_id
            for instance_ref in com_object_instance_refs
            if instance_ref.com_object_ref_id is not None
        }
        com_object_refs: dict[str, ComObjectRef] = {}  # {Id: ComObjectRef}
        com_objects: dict[str, ComObject] = {}  # {Id: ComObject}

        used_module_arguments: dict[str, ModuleDefinitionArgumentInfo] = {
            attribute.ref_id: ModuleDefinitionArgumentInfo()
            for device in devices
            for attribute in device.module_instance_arguments()
        }
        numeric_args: dict[str, ModuleDefinitionNumericArg] = {}
        channels: dict[
            str, ApplicationProgramChannel
        ] = {}  # {Id: ApplicationProgramChannel}
        allocators: dict[str, Allocator] = {}

        with application_program_path.open(mode="rb") as application_xml:
            tree_iterator = ElementTree.iterparse(application_xml, events=("start",))
            # get namespace from root element
            _, elem = next(tree_iterator)
            namespace = elem.tag.split("KNX", maxsplit=1)[0]
            # define namespaced tag strings for faster comparison - ~15% faster
            # than elem.tag.endswith("tagname") or elem.tag == f"{namespace}tagname"
            ns_com_object = f"{namespace}ComObject"
            ns_com_object_ref = f"{namespace}ComObjectRef"
            ns_allocator = f"{namespace}Allocator"
            ns_argument = f"{namespace}Argument"
            ns_numeric_arg = f"{namespace}NumericArg"
            ns_channel = f"{namespace}Channel"
            ns_languages = f"{namespace}Languages"

            for _, elem in tree_iterator:
                if elem.tag == ns_com_object:
                    # we take all since we don't know which are referenced to yet
                    identifier = elem.attrib.get("Id")
                    com_objects[identifier] = ApplicationProgramLoader.parse_com_object(
                        elem, identifier
                    )
                elif elem.tag == ns_com_object_ref:
                    if (_id := elem.attrib.get("Id")) in used_com_object_ref_ids:
                        com_object_refs[_id] = (
                            ApplicationProgramLoader.parse_com_object_ref(elem, _id)
                        )
                    elem.clear()
                elif elem.tag == ns_allocator:  # Allocators/Allocator
                    allocators[elem.attrib.get("Id")] = Allocator(
                        identifier=elem.attrib.get("Id"),
                        name=elem.attrib.get("Name"),
                        start=int(elem.attrib.get("Start")),
                        end=int(elem.attrib.get("maxInclusive")),
                    )
                elif elem.tag == ns_argument:
                    # ModuleDefs/ModuleDef/Arguments/
                    # or ModuleDefs/ModuleDef/SubModuleDefs/ModuleDef/Arguments/
                    if (_id := elem.attrib.get("Id")) in used_module_arguments:
                        allocates = elem.attrib.get("Allocates")
                        used_module_arguments[_id] = ModuleDefinitionArgumentInfo(
                            name=elem.attrib.get("Name"),
                            allocates=int(allocates) if allocates is not None else None,
                        )
                    elem.clear()
                elif elem.tag == ns_numeric_arg:
                    # in dynamic section of Modules
                    if (_id := elem.attrib.get("RefId")) in used_module_arguments:
                        value = elem.attrib.get("Value")
                        numeric_args[_id] = ModuleDefinitionNumericArg(
                            allocator_ref_id=elem.attrib.get("AllocatorRefId"),
                            base_value=elem.attrib.get("BaseValue"),
                            value=int(value) if value is not None else None,
                        )
                    elem.clear()
                elif elem.tag == ns_channel:
                    _id = elem.attrib.get("Id")
                    channels[_id] = ApplicationProgramChannel(
                        identifier=_id,
                        name=elem.attrib.get("Name"),
                        number=elem.attrib.get("Number"),
                        text=elem.attrib.get("Text"),
                        text_parameter_ref_id=elem.attrib.get("TextParameterRefId"),
                    )
                    elem.clear()
                elif elem.tag == ns_languages:
                    elem.clear()
                    # hold iterator for optional translation parsing
                    break
                elem.clear()

            if language_code is not None:
                ApplicationProgramLoader.parse_translations(
                    tree_iterator=tree_iterator,
                    namespace=namespace,
                    com_objects=com_objects,
                    com_object_refs=com_object_refs,
                    used_com_object_ref_ids=used_com_object_ref_ids,
                    channels=channels,
                    language_code=language_code,
                )

            return ApplicationProgram(
                com_objects=com_objects,
                com_object_refs=com_object_refs,
                allocators=allocators,
                module_def_arguments=used_module_arguments,
                numeric_args=numeric_args,
                channels=channels,
            )

    @staticmethod
    def parse_translations(
        tree_iterator: Iterator[tuple[str, Any]],
        namespace: str,
        com_objects: dict[str, ComObject],
        com_object_refs: dict[str, ComObjectRef],
        used_com_object_ref_ids: set[str],
        channels: dict[str, ApplicationProgramChannel],
        language_code: str,
    ) -> None:
        """Parse translations. Replace translated text in com_objects and com_object_refs."""
        _used_com_object_ids = {
            com_object_ref.ref_id for com_object_ref in com_object_refs.values()
        }
        used_translation_ids = (
            _used_com_object_ids | used_com_object_ref_ids | channels.keys()
        )
        in_language = False
        in_translation_ref: str | None = None  # TranslationElement RefId
        # translation_map: {TranslationElement RefId: {AttributeName: Text}}
        translation_map: dict[str, dict[str, str]] = {}

        ns_language = f"{namespace}Language"
        ns_translation_element = f"{namespace}TranslationElement"
        ns_translation = f"{namespace}Translation"

        for _, elem in tree_iterator:
            if elem.tag == ns_language:
                if in_language:
                    # Hitting the next language tag after the one we were looking for.
                    # We don't need anything after that tag (there isn't much anyway)
                    elem.clear()
                    break
                in_language = elem.get("Identifier") == language_code
            elif in_language and elem.tag == ns_translation_element:
                ref_id = elem.get("RefId")
                in_translation_ref = ref_id if ref_id in used_translation_ids else None
            elif (
                in_language
                and in_translation_ref is not None
                and elem.tag == ns_translation
            ):
                translation_map.setdefault(in_translation_ref, {})[
                    elem.get("AttributeName")
                ] = elem.get("Text")
            elem.clear()

        ApplicationProgramLoader.apply_translations(com_object_refs, translation_map)
        ApplicationProgramLoader.apply_translations(com_objects, translation_map)
        ApplicationProgramLoader.apply_translations(channels, translation_map)

    @staticmethod
    def parse_com_object(
        elem: ElementTree.Element,
        identifier: str,
    ) -> ComObject:
        """Parse ComObject tag."""
        return ComObject(
            identifier=identifier,
            name=elem.get("Name"),  # type: ignore[arg-type]
            text=elem.get("Text"),  # type: ignore[arg-type]
            number=int(elem.get("Number", 0)),
            function_text=elem.get("FunctionText"),  # type: ignore[arg-type]
            object_size=elem.get("ObjectSize"),  # type: ignore[arg-type]
            read_flag=parse_xml_flag(elem.get("ReadFlag"), False),
            write_flag=parse_xml_flag(elem.get("WriteFlag"), False),
            communication_flag=parse_xml_flag(elem.get("CommunicationFlag"), False),
            transmit_flag=parse_xml_flag(elem.get("TransmitFlag"), False),
            update_flag=parse_xml_flag(elem.get("UpdateFlag"), False),
            read_on_init_flag=parse_xml_flag(elem.get("ReadOnInitFlag"), False),
            datapoint_types=parse_dpt_types(elem.get("DatapointType")),
            base_number_argument_ref=elem.get("BaseNumber"),
        )

    @staticmethod
    def parse_com_object_ref(
        elem: ElementTree.Element,
        identifier: str,
    ) -> ComObjectRef:
        """Parse ComObjectRef tag."""
        return ComObjectRef(
            identifier=identifier,
            ref_id=elem.get("RefId"),  # type: ignore[arg-type]
            name=elem.get("Name"),
            text=elem.get("Text"),
            function_text=elem.get("FunctionText"),
            object_size=elem.get("ObjectSize"),
            read_flag=parse_xml_flag(elem.get("ReadFlag")),
            write_flag=parse_xml_flag(elem.get("WriteFlag")),
            communication_flag=parse_xml_flag(elem.get("CommunicationFlag")),
            transmit_flag=parse_xml_flag(elem.get("TransmitFlag")),
            update_flag=parse_xml_flag(elem.get("UpdateFlag")),
            read_on_init_flag=parse_xml_flag(elem.get("ReadOnInitFlag")),
            datapoint_types=parse_dpt_types(elem.get("DatapointType")),
            text_parameter_ref_id=elem.get("TextParameterRefId"),
        )

    @staticmethod
    def apply_translations(
        translatable_object_map: dict[str, ComObject]
        | dict[str, ComObjectRef]
        | dict[str, ApplicationProgramChannel],
        translation_map: dict[str, dict[str, str]],
    ) -> None:
        """Apply translations to Objects."""
        for identifier in translatable_object_map.keys() & translation_map.keys():
            translation = translation_map[identifier]
            obj = translatable_object_map[identifier]
            if _text := translation.get("Text"):
                obj.text = _text
            if hasattr(obj, "function_text") and (
                _function_text := translation.get("FunctionText")
            ):
                obj.function_text = _function_text

    @staticmethod
    def get_application_program_files_for_devices(
        devices: list[DeviceInstance],
    ) -> dict[str, list[DeviceInstance]]:
        """Do not load the same application program multiple times."""
        result: dict[str, list[DeviceInstance]] = {}
        for device in devices:
            if device.application_program_ref:
                # zipfile.Path hashes are not equal, therefore we use str to create the struct
                xml_file_name = device.application_program_xml()
                result.setdefault(xml_file_name, []).append(device)
        return result
